#!/usr/bin/python
# -*- coding: utf-8 -*-
import os
import sys
import time
import socket
import urllib.parse
import configparser

from io import StringIO

from libraries.exceptions.auraexceptions import LQConnectionError


""" LiquidSoapClient Class 
    Repräsentiert alle Kommandos, die Soundserver und Recorder kennen
"""


class LiquidSoapClient:
    debug = False
    socket_path = ""

    def __init__(self, config, socket_filename):
        """
        Constructor
        @type    socket_path: string
        @param   socket_path: Der Pfad zum Socket des Liquidsoap-Scripts
        """

        self.socket_path = config.get('socketdir') + '/' + socket_filename
        self.debug = config.get("debug")

        if self.debug > 2:
            print("socketpath: " + self.socket_path)

        # init
        self.connected = False
        self.can_connect = True
        self.message = ''
        self.client = None
        if sys.version_info <= (3, 2):
            self.metareader = configparser.ConfigParser({'strict': False, 'interpolation': None})
        else:
            self.metareader = configparser.ConfigParser()

    # ------------------------------------------------------------------------------------------ #
    def connect(self):
        """
        Verbindung herstellen
        """
        try:
            self.client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.client.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.client.connect(self.socket_path)
        except socket.error as e:
            self.connected = False
            print(self.socket_path)
            raise LQConnectionError(e, self.socket_path)
            return False
        else:
            self.can_connect = True
            self.connected = True
            return True

    # ------------------------------------------------------------------------------------------ #
    def is_connected(self):
        return self.connected

    # ------------------------------------------------------------------------------------------ #
    def write(self, data):
        """
        Auf den Socket schreiben
        @type    data: string
        @param   data: Der String der gesendet wird
        """
        if self.connected:
            self.client.sendall(data.decode("UTF-8"))

    # ------------------------------------------------------------------------------------------ #
    def read_all(self, timeout=2):
        """
        Vom Socket lesen, bis dieser "END" sendet
        @type    timeout: int
        @param   timeout: Ein optionales Timeout
        @rtype:  string
        @return: Die Antwort des Liquidsoap-Servers
        """
        # make socket non blocking
        # self.client.setblocking(0)

        data = '';
        self.client.settimeout(3)

        # recv something
        # while True:
        try:
            time.sleep(0.1)
            raw = self.client.recv(8192)
            data = raw.decode()
        except Exception as e:
            print(e)
            pass

        return data

    # ------------------------------------------------------------------------------------------ #
    def read(self):
        """
        read from socket and store return value in self.message
        @rtype:  string
        @return: The answer of liquidsoap server
        """
        if self.connected:
            ret = self.read_all().splitlines()

            try:
                last = ret.pop()
                if last == 'Bye!':
                    ret.append(last)
                elif last != "END":
                    print("got no end from liquidsoap")

                self.message = str.join("\r\n", ret)

            except Exception as e:
                print(e)

            return self.message

    # ------------------------------------------------------------------------------------------ #
    def close(self):
        """
        Quit senden und Verbindung schließen
        """
        if self.connected:
            message = "quit\r"
            self.client.sendall(message.decode("UTF-8"))
            self.client.close()
            self.connected = False

    # ------------------------------------------------------------------------------------------ #
    def command(self, namespace, command, param=""):
        """
        Kommando an Liquidosap senden
        @type    command:   string
        @param   command:   Kommando
        @type    namespace: string
        @param   namespace: Namespace/Kanal der angesprochen wird
        @type    param:     mixed
        @param   param:     ein optionaler Parameter
        @rtype:  string
        @return: Die Antwort des Liquidsoap-Servers
        """

        param = (param.strip() if param.strip() == "" else " " + urllib.parse.unquote(param.strip()))
        if self.connected:
            # print namespace + '.' + command + param + "\n"
            if namespace is "":
                message = str(command) + str(param) + str('\n')
            else:
                message = str(namespace) + str('.') + str(command) + str(param) + str('\n')

            try:
                if self.debug > 2:
                    print("LiquidSoapClient sending to LiquidSoap Server: "+message[0:len(message)-1])

                # send all the stuff over the socket to liquidsoap server
                self.client.sendall(message.encode())

                if self.debug > 2:
                    print("LiquidSoapClient waiting for reply from LiquidSoap Server")

                # wait for reply
                self.read()

                if self.debug > 2:
                    print("LiquidSoapClient got reply")
                # self.client.close()
            except:
                print("Unexpected error:", sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
                raise

            return self.message
        else:
            if self.debug:
                print("LiquidSoapClient not connected to LiquidSoap Server")
            return False

    # ------------------------------------------------------------------------------------------ #
    def simplecommand(self, command):
        """
        Parameterloses Kommando ohne Namespace senden
        @type    command:   string
        @param   command:   Kommando
        @rtype:  string
        @return: Die Antwort des Liquidsoap-Servers
        """

        if self.connected:
            message = str(command) + '\n'
            self.client.sendall(message.decode("UTF-8"))
            self.read()
            # self.client.close()
            return self.message

    # ------------------------------------------------------------------------------------------ #
    def getMetadata(self, rid):
        """
        Parameterloses Kommando ohne Namespace senden
        @type    rid:   string/int
        @param   rid:   Die ID eines Requests
        @rtype:  dict
        @return: Die Metadaten als dict
        """
        meta = self.command('metadata ' + str(rid), 'request')

        meta = '[root]\n' + meta
        if sys.version_info <= (3, 2):
            meta = StringIO.StringIO(meta)
            try:
                self.metareader.readfp(meta)
            except configparser.ParsingError:
                return False
        else:
            try:
                self.metareader.read_string(meta)
            except configparser.ParsingError:
                return False
        return self.metareader

    # ------------------------------------------------------------------------------------------ #
    def help(self):
        """
        get liquidsoap server help
        @rtype:  string
        @return: the response of the liquidsoap server
        """
        if self.connected:
            self.command('help', '')
            print('read (' + str(len(self.message)) + ' bytes): ' + self.message)
            return self.message

    # ------------------------------------------------------------------------------------------ #
    def version(self):
        """
        Liquidsoap get version
        @rtype:  string
        @return: the response of the liquidsoap server
        """
        if self.connected:
            message = 'version'
            self.command(message, '')
            return self.message

    # ------------------------------------------------------------------------------------------ #
    def uptime(self):
        """
        Liquidsoap get uptime
        @rtype:  string
        @return: Die Antwort des Liquidsoap-Servers
        """

        if self.connected:
            self.command('uptime', '')
            print('read (' + str(len(self.message)) + ' bytes): ' + self.message)
            return self.message

    # ------------------------------------------------------------------------------------------ #
    def byebye(self):
        """
        Liquidsoap say byebye
        @rtype:  string
        @return: Die Antwort des Liquidsoap-Servers
        """

        if self.connected:
            message = 'quit'
            self.command("", message)  # client.sendall(message.encode())
            # self.read()
            # print("read: " + self.message)
            return self.message