import os
import codecs
import urllib
import tempfile
import simplejson
import traceback
import datetime

from libraries.base.parsexml import parsexml
from modules.communication.liquidsoap.playerclient import LiquidSoapPlayerClient
from modules.communication.liquidsoap.recorderclient import LiquidSoapRecorderClient
from modules.communication.liquidsoap.initthread import LiquidSoapInitThread
from libraries.enum.consolecolor import TerminalColors
from libraries.exceptions.auraexceptions import LQConnectionError
from libraries.database.broadcasts import TrackService
from libraries.exceptions.exception_logger import ExceptionLogger
from libraries.enum.scheduleentrytypes import ScheduleEntryTypes


class LiquidSoapCommunicator(ExceptionLogger):
    lqc = None
    lqcr = None
    debug = False
    transaction = 0
    # controller = None
    channels = None
    scheduler = None

    # ------------------------------------------------------------------------------------------ #
    def __init__(self, config):
        """
        Constructor
        @type    lqs_socket: string
        @param   lqs_socket: Liquidsoap Player Socket
        @type    lqs_recsocket: string
        @param   lqs_recsocket: Liquidsoap Recorder Socket
        """
        self.config = config
        self.debug = config.get("debug")
        self.lqc = LiquidSoapPlayerClient(config, "simplestmixer.sock")
        self.lqcr = LiquidSoapRecorderClient(config, "record.sock")

        #try:
        #    self.get_all_channels()
        #except LQConnectionError as lqe:
        #    print(lqe)

    #def set_controller(self, controller):
    #    self.controller = controller



    # ------------------------------------------------------------------------------------------ #
    def set_volume(self, mixernumber, volume):
        print("LiquidSoapCommunicator.set_volume")
        return self.__send_lqc_command__(self.lqc, "mixer", "volume", mixernumber, volume)

    # ------------------------------------------------------------------------------------------ #
    def get_active_mixer(self):
        """
        get active mixer in liquidsoap server
        :return:
        """
        activeinputs = []

        # enable more control over the connection
        self.enable_transaction(True)

        inputs = self.get_all_channels()

        cnt = 0
        for input in inputs:
            status = self.__get_mixer_status__(cnt)

            if "selected=true" in status:
                activeinputs.append(input)

            cnt = cnt + 1

        self.disable_transaction(True)

        return activeinputs

    # ------------------------------------------------------------------------------------------ #
    def get_active_channel(self):
        """
        gets active channel from programme
        :return:
        """
        active_entry = self.scheduler.get_active_entry()
        if active_entry is None:
            return ""
        return active_entry.type

    # ------------------------------------------------------------------------------------------ #
    def get_mixer_status(self):
        inputstate = {}

        self.enable_transaction()
        self.open_conn(self.lqc)

        inputs = self.get_all_channels()

        cnt = 0
        for input in inputs:
            inputstate[input] = self.__get_mixer_status__(cnt)
            cnt = cnt + 1

        self.disable_transaction()
        self.close_conn(self.lqc)

        return inputstate

    # ------------------------------------------------------------------------------------------ #
    def activate(self, entry):
#        active_source = self.scheduler.get_active_entry().type
#        print(active_source)
        if self.scheduler.get_active_entry().type == entry.type:
            print(TerminalColors.OK_PINK.value + "LiquidSoapCommunicator " + entry.type.value + " already active! Time: " + str(datetime.datetime.now()) + TerminalColors.ENDC.value)
            self.enable_transaction(True)
            if entry.type == ScheduleEntryTypes.FILESYSTEM:
                self.playlist_push(entry.source)
            if entry.type == ScheduleEntryTypes.HTTP:
                self.set_http_url(entry.source)
            self.disable_transaction(True)

            self.insert_track_service_entry(entry)

        else:
            print(TerminalColors.OK_PINK.value + "LiquidSoapCommunicator is activating " + entry.type.value + "! Time: " + str(datetime.datetime.now()) + TerminalColors.ENDC.value)

            self.enable_transaction(True)

            if entry.type == ScheduleEntryTypes.FILESYSTEM:
                self.playlist_push(entry.source)
            if entry.type == ScheduleEntryTypes.HTTP:
                self.set_http_url(entry.source)

            others = self.all_inputs_but(entry.type.value)
            for o in others:
                self.channel_volume(o, 0)
            self.channel_volume(entry.type.value, entry.volume)
            self.disable_transaction(True)

            self.insert_track_service_entry(entry)
            #self.playlist_insert()

    # ------------------------------------------------------------------------------------------ #
    def insert_track_service_entry(self, schedule_entry):
        trackservice_entry = TrackService()
        trackservice_entry.playlist_id = schedule_entry.playlist_id
        trackservice_entry.entry_num = schedule_entry.entry_num
        trackservice_entry.source = schedule_entry.source

        trackservice_entry.store(add=True, commit=True)

    # ------------------------------------------------------------------------------------------ #
    def all_inputs_but(self, input_type):
        activemixer_copy = self.get_all_channels().copy()

        try:
            activemixer_copy.remove(input_type)
        except ValueError:
            # Value not in list
            return activemixer_copy

        return activemixer_copy

    # ------------------------------------------------------------------------------------------ #
    def get_all_channels(self):
        if self.channels is None or self.channels == ['']:
            self.channels = self.__send_lqc_command__(self.lqc, "mixer", "inputs")

        return self.channels

    # ------------------------------------------------------------------------------------------ #
    def reload_channels(self):
        self.channels = None
        return self.get_all_channels()

    # ------------------------------------------------------------------------------------------ #
    def __get_mixer_status__(self, mixernumber):
        return self.__send_lqc_command__(self.lqc, "mixer", "status", mixernumber)

    # ------------------------------------------------------------------------------------------ #
    def init_player(self):
        LiquidSoapInitThread(self).start()
        return "Started LiquidSoapInitThread!"

    # ------------------------------------------------------------------------------------------ #
    def get_client(self):
        return self.lqc

    # ------------------------------------------------------------------------------------------ #
    def all_data(self):
        """
        Gibt Metadaten aller Kanäle als JSON-String an den Client zurück
        @rtype:   string/None
        @return:  Die Antwort des Liquidsoap-Servers

        """
        channels = self.__send_lqc_command__(self.lqc, 'list_channels')

        if not isinstance(channels, list):
            self.warning('01')
            self.notifyClient()
            return

        data = {}
        pdata = {}

        try:
            self.is_intern = True
            playlist_data = simplejson.loads(self.playlist_data(True))
            self.is_intern = False
        except:
            self.warning('01')
            self.notifyClient()
            return

        # Status des Playlistkanals abfragen
        status = self.__send_lqc_command__(self.lqc, 'status', 'mixer', '0')

        states = status.split(' ')
        state_data = {}

        # Die Stati in python dicts einlesen
        for state in states:
            item = state.split('=')
            try:
                state_data[item[0]] = item[1]
            except:
                self.warning('01')
                self.notifyClient()
                return

        remaining = self.__send_lqc_command__(self.lqc, 'playlist_remaining')
        state_data['remaining'] = remaining
        # Die Metadaten der Playlist
        pdata['state'] = state_data
        pdata['tracks'] = playlist_data
        data['playlist'] = pdata

        # Servermeldungen abschalten
        self.is_intern = True
        # die channel queues einlesen

        for channel in channels:
            data[channel] = self.channel_queue(channel, True)
        # Servermeldungen einschalten
        self.is_intern = False
        self.success('00', data)
        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def ping(self):
        """
        dem Client antworten
        """
        return self.message('OK')

    # ------------------------------------------------------------------------------------------ #
    def channel_insert(self, channel, uri, pos):
        """
        Track in einen Channel einfuegen
        @type     channel: string
        @param    channel: Kanal
        @type     uri:     string
        @param    uri:     Uri - z.B. file:///my/audio/mp3
        @type     pos:     int
        @param    pos:     Die Position an der eingefügt werden soll
        @rtype:   string/None
        @return:  Die Antwort des Liquidsoap-Servers
        """
        message = self.__send_lqc_command__(self.lqc, 'insert', uri, pos, channel)
        message = message.strip()

        try:
            if int(message) > -1:
                self.success()
                return self.message(message)
        except:
            self.warning('01')
            self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def channel_move(self, channel, fromPos, toPos):
        """
        Channel-Eintrag von Position fromPos nach Position toPos verschieben
        @type     channel: string
        @param    channel: Kanal
        @type     fromPos: int
        @param    fromPos: die Position des Eintrags, der verschoben wird
        @type     toPos:   int
        @param    toPos:   Zielposition
        @rtype:   string
        @return:  Die Antwort des Liquidsoap-Servers
        """
        message = self.__send_lqc_command__(self.lqc, 'get_queue', channel, 'secondary_queue')

        rids = message.strip().split(' ')

        try:
            rid = rids[int(fromPos) - 1]
        except:
            self.warning('01')
            self.notifyClient()
            return
        try:
            target = rids[int(toPos) - 1]
        except:
            self.warning('01')
            self.notifyClient()
            return

        if rids[int(fromPos) - 1] == rids[int(toPos) - 1]:
            self.warning('02')
            self.notifyClient()
            return

        message = self.__send_lqc_command__(self.lqc, 'move', rid, str(int(toPos) - 1), channel)
        message = message.strip()

        if message.strip().find('OK') > -1:
            self.success()
            self.notifyClient()
            return
        else:
            self.warning('03')
            self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def channel_off(self, channel):
        """
        Channel deaktivieren
        @type     channel: string
        @param    channel: Kanal
        @rtype:   string
        @return:  Die Antwort des Liquidsoap-Servers
        """
        # internal channel name for playlist is 'common'
        if channel == 'playlist':
            channel = 'common'

        channels = self.__send_lqc_command__(self.lqc, 'list_channels', False)

        index = channels.index(channel)
        message = self.__send_lqc_command__(self.lqc, 'deactivate', str(index))
        if message.find('selected=false'):
            self.success()
        else:
            self.warning('01')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def channel_on(self, channel):
        """
        Channel aktivieren
        @type     channel: string
        @param    channel: Kanal
        @rtype:   string
        @return:  Die Antwort des Liquidsoap-Servers
        """

        # Find channels
        if channel == 'playlist':
            channel = 'common'

        channels = self.__send_lqc_command__(self.lqc, 'list_channels', False)

        index = channels.index(channel)
        # a activate channel
        message = self.__send_lqc_command__(self.lqc, 'activate', str(index))

        if message.find('selected=true'):
            self.success()
        else:
            self.warning('01')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def channel_queue(self, channel, raw=False):
        """
        Channel Queue abrufen
        @type     channel: string
        @param    channel: Kanal
        @type     raw:     boolean
        @param    raw:     Wenn true, Rückgabe als Python dict Object, andernfalls als JSON-String
        @rtype:   string/dict
        @return:  Der Channel Queue
        """
        data = {}

        # queue will return request id's (rids)

        message = self.__send_lqc_command__(self.lqc, 'get_queue', channel)

        rids = message.strip().split(' ')
        data['tracks'] = []
        for rid in rids:
            if rid != '':
                # get each rids metadata
                metadata = self.__send_lqc_command__(self.lqc, 'getMetadata', rid)
                track = self._metadata_format(metadata)
                if not 'title' in track:
                    if 'location' in track:
                        track['title'] = os.path.basename(track['location'])
                    elif 'filename' in track:
                        track['title'] = os.path.basename(track['filename'])
                    else:
                        track['title'] = 'unknown'

                data['tracks'].extend([track])
        channels = self.__send_lqc_command__(self.lqc, 'list_channels')

        """
        now get channels state 
        self.lqc.status: ready=false volume=100% single=false selected=false remaining=0.00
        """
        try:
            index = channels.index(channel)
            status = self.__send_lqc_command__(self.lqc, 'status', 'mixer', str(index + 1))
            states = status.split(' ')
            state_data = {}
            for state in states:
                item = state.split('=')
                if len(item) > 1:
                    state_data[item[0]] = item[1]
        except:
            state_data = {}
            self.error('01')
            self.notifyClient()
            return

        data['state'] = state_data

        if raw:
            # return the list internal
            data['state'] = state_data
            return data
        else:
            self.success('00', data)
            self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def channel_remove(self, channel, pos):
        """
        Channel-Eintrag löschen
        @type     channel: string
        @param    channel: Kanal
        @type     pos:     int
        @param    pos:     Position des Eintrags
        """
        # Es kann nur vom Secondary Queue gelöscht werden
        # Falls der Track im Primary Queue gelöscht werden soll, ist ein skip nötg

        message = self.__send_lqc_command__(self.lqc, 'get_queue', channel, 'secondary_queue')
        rids = message.strip().split(' ')
        try:
            rid = rids[int(pos) - 1]
        except:
            self.warning('02')
            self.notifyClient()
            return
        message = self.__send_lqc_command__(self.lqc, 'remove', rid, channel)
        if message.find('OK') > -1:
            self.success()
        else:
            self.warning('01')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def channel_seek(self, channel, duration):
        """
        Im aktuell spielenden Track auf dem Kanal <channel> <duration> Sekunden "vorspulen"
        @type     channel:  string
        @param    channel:  Kanal
        @type     duration: int
        @param    duration: Dauer in Sekunden
        """
        # Liquidsoap Kommando
        data = self.__send_lqc_command__(self.lqc, 'seek', duration, channel)

        # Resultate prüfen
        if self._check_result(data):
            self.success('00', self.lq_error['value'])
        else:
            self.warning('01')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def channel_skip(self, channel):
        """
        Kanal skippen
        @type     channel:  string
        @param    channel:  Kanal
        """

        # Liquidsoap Kommando
        channels = self.__send_lqc_command__(self.lqc, 'list_channels')

        foundChannel = ''
        if not isinstance(channels, list):
            self.error('02')
        else:
            for index, item in enumerate(channels):
                if item == channel:
                    foundChannel = self.__send_lqc_command__(self.lqc, 'skip', 'mixer', str(index + 1))
                    break

            if foundChannel.strip().find('OK') > -1:
                self.success()
            elif len(channels) < 1:
                self.warning('01')
            else:
                self.error('03')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def channel_activate(self, channel, activate):
        channels = self.get_all_channels()

        try:
            index = channels.index(channel)
            if len(channel) < 1:
                self.warning('02')
        except:
            self.error('03')

        else:
            message = self.__send_lqc_command__(self.lqc, "mixer", "select", index, activate)
            return message


    # ------------------------------------------------------------------------------------------ #
    def channel_volume(self, channel, volume):
        """
        Lautstärke auf Kanal <channel> setzen
        @type     channel:  string
        @param    channel:  Kanal
        @type     volume:   int
        @param    volume:  Lautstärke von 1-100
        """
        # Liquidsoap Kommando
        #channels = self.__sendLqcCommand__(self.lqc, 'list_channels', False)
        channels = self.get_all_channels()

        try:
            index = channels.index(channel)
            if len(channel) < 1:
                self.warning('02')
        except:
            self.error('03')

        else:
            message = self.__send_lqc_command__(self.lqc, 'mixer', 'volume', str(index), str(int(volume)))
            return message

#            if message.find('volume=' + str(volume) + '%'):
#                self.controller.success('01', str(volume))
#            else:
#                self.controller.warning('01')

    # ------------------------------------------------------------------------------------------ #
    def current_data(self):
        """
        Metadaten des gespielten Tracks im JSON-Format
        Beispiel: {"title": "Deserted Cities of the Heart", "filename": "/home/michel/Nas-audio/cream/the_very_best_of/17_Deserted_Cities_of_the_Heart.mp3", "source": "ch2", "on_air": "2014/07/23 23:46:37",  "rid": "2"}
        """

        # Liquidsoap Kommando
        message = self.__send_lqc_command__(self.lqc, 'currentTrack')

        rid = message.strip()

        metadata = self.__send_lqc_command__(self.lqc, 'getMetadata', rid)

        data = self._metadata_format(metadata)

        if data:
            self.success('00', simplejson.dumps(data))
        elif rid == '':
            self.warning('01')
        else:
            self.warning('02', rid)

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def get_channel_state(self, channel):

        if channel == 'playlist':
            channel = 'common'

        channels = self.__send_lqc_command__(self.lqc, 'list_channels', False)

        index = channels.index(channel)
        state_data = {}
        try:
            index = channels.index(channel)
            status = self.__send_lqc_command__(self.lqc, 'status', 'mixer', str(index + 1))
            states = status.split(' ')
            for state in states:
                item = state.split('=')
                if len(item) > 1:
                    state_data[item[0]] = item[1]
        except:
            state_data = {}
            self.error('01')
            self.notifyClient()
            return
        self.success('00', simplejson.dumps(state_data))
        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def help(self):
        """
        Gibt die Hilfe aus
        """

        errNum = '11'
        try:
            file = open(os.path.dirname(os.path.abspath(__file__)) + '/doc/comba.hlp', 'r')
            doc = file.read()
            return self.message(doc)
        except:
            self.warning('01')
            self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def liquidsoap_help(self):
        data = self.__send_lqc_command__(self.lqc, 'help')
        if not data:
            self.warning('01')
        else:
            self.success('00', data)
        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def list_channels(self):
        """
        Channels auflisten (Simple JSON)
        """
        # Liquidsoap Kommando

        channels = self.__send_lqc_command__(self.lqc, 'list_channels')

        if not isinstance(channels, list):
            self.error('02')
        elif len(channels) < 1:
            self.warning('01')
        else:
            self.success('00', channels)

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def playlist_data(self, raw=False):
        """
        Aktuelle Playlist Daten im JSON-Format
        """

        # Liquidsoap Kommando
        data = self.__send_lqc_command__(self.lqc, 'playlistData')
        if not raw:
            self.success('00', simplejson.loads(data))
            self.notifyClient()
        else:
            return data

    # ------------------------------------------------------------------------------------------ #
    def playlist_flush(self):
        """
        Aktuelle Playlist leeren
        """
        data = self.__send_lqc_command__(self.lqc, 'flush')
        self.success('00')
        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def playlist_insert(self, uri, pos):
        """
        Track in die Playlist einfuegen
        """
        data = self.__send_lqc_command__(self.lqc, 'insert', uri, pos)
        if not self._check_result(data):
            self.warning('01')
        else:
            self.success('00')
        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def playlist_load(self, uri):
        """
        Playlist laden
        @type   uri:   string
        @param  uri:   Uri der Playlist
        """
        try:
            xml = urllib.urlopen(uri).read().decode('utf8')

        except:
            try:
                xml = open(uri).read().decode('utf8')
            except:
                self.error('01', self.lq_error['message'])
                self.notifyClient()
                return

        (num, filename) = tempfile.mkstemp(suffix=".xspf")

        with codecs.open(filename, "w", encoding='utf8') as text_file:
            text_file.write(xml)

        playlist = parsexml(xml)

        if not isinstance(playlist, dict):
            self.error('02')
            self.notifyClient()
        else:
            self.__send_lqc_command__(self.lqc, 'flush')
            data = self.__send_lqc_command__(self.lqc, 'loadPlaylist', filename)

            if not self._check_result(data):
                self.error('01', self.lq_error['message'])

            else:
                os.remove(filename)
                self._updateEventQueue(playlist)
                event = {'job': 'loadplaylist', 'uri': uri}
                self.messenger.fire_event('loadplaylist', event, 'player')
                self.success('00')

            self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def playlist_move(self, fromPos, toPos):
        """
        Playlist-Eintrag von Position fromPos nach Position toPos verschieben
        @type     fromPos: int
        @param    fromPos: die Position des Eintrags, der verschoben wird
        @type     toPos:   int
        @param    toPos:   Zielposition
        """
        data = self.__send_lqc_command__(self.lqc, 'move', str(int(fromPos) + 1), str(int(toPos) + 1))

        if not self._check_result(data):
            self.warning('01')
        else:
            self.success('00')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def playlist_pause(self):
        """
        Playlist pausieren
        """
        data = self.__send_lqc_command__(self.lqc, 'pause')

        if not self._check_result(data):
            self.info('01')
        else:
            self.success('00')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def playlist_stop(self):
        """
        Playlist stoppen - der Kanal wird deaktiviert
        """
        # Kanal 0 (Playlist) deaktivieren
        self.__send_lqc_command__(self.lqc, 'deactivate', '0')

        data = self.__send_lqc_command__(self.lqc, 'pause')

        if not self._check_result(data):
            self.info('01')
        else:
            self.success('00')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def playlist_play(self, when='now'):
        """
        Playlist starten
        @type   when:   string
        @param  when:   Wenn "now" werden alle anderen Kanäle deaktiviert und geskipped
        """
        # Playlist Kanal aktivieren
        self.__send_lqc_command__(self.lqc, 'activate', '0')

        if when == 'now':
            # immediately skip all playing channels
            # and activate the playlist channel
            channels = self.__send_lqc_command__(self.lqc, 'list_channels')
            if not isinstance(channels, list):
                self.error('03')
            elif len(channels) < 1:
                self.warning('02')
            else:
                # xrange
                for i in range(len(channels)):
                    status = self.__send_lqc_command__(self.lqc, 'status', 'mixer', str(i + 1))
                    if "selected=true" in status:
                        status = self.__send_lqc_command__(self.lqc, 'deactivate', str(i + 1))
                        status = self.__send_lqc_command__(self.lqc, 'skip', 'mixer', str(i + 1))
                        self.__send_lqc_command__(self.lqc, 'activate', '0')

        # send the play command
        data = self.__send_lqc_command__(self.lqc, 'play')
        if not self._check_result(data):
            self.info('01')
        else:
            self.success('00')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def set_http_url(self, uri):
        return self.__send_lqc_command__(self.lqc, 'http', 'url', uri)

    # ------------------------------------------------------------------------------------------ #
    def playlist_push(self, uri):
        """
        Eine Uri in die Playlist einfügen
        @type   uri:   str
        @param  uri:   Die Uri
        """
        return self.__send_lqc_command__(self.lqc, 'fs', 'push', uri)
#        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def playlist_remove(self, pos):
        """
        Playlist-Eintrag löschen
        @type     pos:     int
        @param    pos:     Position des Eintrags
        """
        data = self.__send_lqc_command__(self.lqc, 'remove', pos)

        if not self._check_result(data):
            self.info('01')
        else:
            self.success('00')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def playlist_seek(self, duration):
        """
        Im aktuell spielenden Track auf dem der Playlist "vorspulen"
        @type     duration: int
        @param    duration: Dauer in Sekunden
        """
        data = self.__send_lqc_command__(self.lqc, 'seek', duration)

        # Resultate prüfen
        if self._check_result(data):
            self.success('00', self.lq_error['value'])
        else:
            self.warning('01')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def playlist_skip(self):
        """
        Playlist skippen
        """
        data = self.__send_lqc_command__(self.lqc, 'skip')

        self.success('00')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def version(self):
        """
        get version
        """
        data = self.__send_lqc_command__(self.lqc, 'version')
        self.success('00', data)
        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def uptime(self):
        """
        get uptime
        """
        data = self.__send_lqc_command__(self.lqc, 'uptime')
        self.success('00', data)
        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def on_air(self):
        """
        get whats playing now
        """
        data = self.__send_lqc_command__(self.lqc, 'on_air')
        self.success('00', data)
        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def recorder_start(self):
        """
        Recorder starten
        @rtype:   string
        @return:  Die Antwort des Liquidsoap-Servers
        """
        message = self.__send_lqc_command__(self.lqcr, 'start_record')
        if message.strip() == 'OK':
            self.success('00')
        else:
            self.warning('01')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def recorder_stop(self):
        """
        Recorder stoppen
        """
        message = self.__send_lqc_command__(self.lqcr, 'stop_record')

        if message.strip() == 'OK':
            self.success('00')
        else:
            self.warning('01')

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def recorder_data(self):
        """
        Status-Daten des Recorders
        Rückgabe-Beispiel: /var/audio/rec/2014-05-13/2014-05-13-22-00.wav,30 - Aufnahme von 30% der angegebenen Audiodatei
        """
        message = self.__send_lqc_command__(self.lqcr, 'recorder_data')
        l = message.split(',')
        data = {}

        if not isinstance(l, list):
            data = {'file': '', 'recorded': ''}
            self.warning('01')
        else:
            data['file'] = l[0]
            if len(l) > 1:
                data['recorded'] = l[1]
            else:
                data['recorded'] = ''
            self.success('00', data)

        self.notifyClient()

    # ------------------------------------------------------------------------------------------ #
    def __send_lqc_command__(self, lqs_instance, namespace, command, *args):
        """
        Ein Kommando an Liquidsoap senden
        @type  lqs_instance: object
        @param lqs_instance: Instance of LiquidSoap Client
        @type  namespace:    string
        @param namespace:    Namespace of function
        @type  command:      string
        @param command:      Function name
        @type args:          list
        @param args:         List of parameters
        @rtype:              string
        @return:             Response from LiquidSoap
        """
        try:
            # connect
            if not self.transaction:
                self.open_conn(lqs_instance)

            if self.debug:
                print("LiquidSoapCommunicator is calling " + str(namespace) + "." + str(command) + str(args))

            # call wanted function ...
            func = getattr(lqs_instance, namespace)
            # ... and fetch the result
            result = func(command, *args)

            if self.debug:
                print("LiquidSoapCommunicator got response " + str(result))

            # say byebye
            if not self.transaction:
                self.close_conn(lqs_instance)

        except Exception as e:
            # Verbindung gescheitert - Fehler an Client
            #            if command.find('record') > -1:
            #                raise e
            # self.fatal('02')
            #            else:
            #                raise e
            # self.fatal('01')
            traceback.print_exc()
            raise e

            # Instanz/Thread zerstören - aufrufende Funktion wird nicht weiter abgearbeitet
            #del self
        else:
            return result

    # ------------------------------------------------------------------------------------------ #
    def enable_transaction(self, connect=False):
        self.transaction = self.transaction + 1

        if self.debug:
            print(TerminalColors.WARNING.value + "ENabling transaction! cnt: " + str(self.transaction) + TerminalColors.ENDC.value)

        if self.transaction > 1:
            return

        if connect:
            self.open_conn()

    # ------------------------------------------------------------------------------------------ #
    def disable_transaction(self, disconnect=False):
        self.transaction = self.transaction - 1

        if self.debug:
            print(TerminalColors.WARNING.value + "DISabling transaction! cnt: " + str(self.transaction) + TerminalColors.ENDC.value)

        if self.transaction > 0:
            return

        if disconnect:
            self.close_conn()

        self.transaction = False

    # ------------------------------------------------------------------------------------------ #
    def open_conn(self, socket=None):
        if socket is None:
            socket = self.lqc

        if self.transaction > 1:
            return # already connected

        if self.debug:
            print(TerminalColors.OK_GREEN.value + "LiquidSoapCommunicator opening conn" + TerminalColors.ENDC.value)

        socket.connect()

    # ------------------------------------------------------------------------------------------ #
    def close_conn(self, socket=None):
        if socket is None:
            socket = self.lqc

        if self.transaction > 0:
            return # still need conn

        socket.byebye()

        if self.debug:
            print(TerminalColors.OK_BLUE.value + "LiquidSoapCommunicator closed conn" + TerminalColors.ENDC.value)
